import numpy as np
from paratc import _utils, _const, tctools
from paratc.tc_models import TCModel

class Holland1980( TCModel ):
    '''
    Class and functions for generating parametric tropical cyclones according to
    (Holland, 1980). Can be used either as a library to access pressure and wind
    functions, or as an instance of TCModel to automatically handle storm 
    generation.

    Initialize a Holland1980 tropical cyclone model. 
        
    Args:
        track (pd.DataFrame): Dataframe whose columns are track parameters. 
            This cyclone model requires at least: 
                'lon' : Longitude in degrees
                'lat' : Latitude in degrees
                'rmw' : Radius of max winds in km
                'pcen' : Central pressure in mb
                'penv' : Environmental pressure in mb
            Other columns may be required depending on B_model used. If rmw is
            missing, it will be generated using a statistical model. If 'B' is a
            columns in track, this will be used.
        grid_lon (np.ndarray): 2D grid longitudes
        grid_lat (np.ndarray): 2D grid latitudes
        B_model (str): The B_model to use. Default is powell05.
        B_min (float): Minimum value for B shape parameter.
        B_max (float): Maximum value for B shape parameter.
        **kwargs: Any extra keyword arguments are passed to TCModel()

    Attributes:
        data (xr.DataSet): Contains generated storm parameters such as windspeed
            and pressure. The grid is generated using the grid_lon and grid_lat
            parameters provided by the user. When first called, this class will
            generate pressure and gradient wind fields. The wind fields are
            unadjusted by background flow and winds are perpendicular to the
            storm center.
        track (pd.Dataframe): Processed track information. User input is stored
            to the instance and some extra columns are calculated (utrans, vtrans
            B, rmw, timestep).
    '''

    def __init__(self, track, grid_lon, grid_lat,
                 B_model = 'powell05',
                 B_min = 1, B_max = 2.5, 
                 **kwargs):
        
        # Check general parameters and make grid dataset
        super().__init__(track, grid_lon, grid_lat, **kwargs)
        data = self.data
        track = self.track
    
        # Make B if it isn't in track dataframe
        if B_model == 'powell05':
            track['B'] = self.B_powell05( track.rmw, track.lat )
        elif B_model == 'wr04':
            track['B'] = self.B_wr04( track.rmw, track.vmax, track.lat )
        elif B_model == 'vickery00':
            track['B'] = self.B_vickery00( track.pdelta, track.rmw )
        elif B_model == 'holland80':
            track['B'] = self.B_holland80( track.pdelta, track.vmax )
        elif B_model is None and 'B' not in track:
            raise Exception(' Expected to find B column in track since B_model = None')
        else:
            raise Exception(f'B_model unknown: {B_model}')
        track['B'] = np.clip( track['B'], B_min, B_max )

        # Generate pressure and gradient wind speeds -- add penv to initial pressure
        n_time = len(track)
        pressure = np.zeros( (n_time, *grid_lon.shape) )
        pressure[0] = track.penv[0]
        wind_g = np.zeros( (n_time, *grid_lon.shape) )
    
        for tii in range( 1, n_time ):
            tr_ii = track.iloc[tii]
    
            # Make pressure and gradient wind
            dist_ii = data.dist_cent[tii].values
            pressure[tii] = self.pressure_equation( dist_ii, 
                                                    rmw = tr_ii.rmw, 
                                                    B = tr_ii.B, 
                                                    penv = tr_ii.penv,
                                                    pcen = tr_ii.pcen, 
                                                    lat = tr_ii.lat )
            wind_g[tii] = self.gradient_wind_equation( dist_ii, 
                                                       rmw = tr_ii.rmw, 
                                                       B = tr_ii.B, 
                                                       pdelta = tr_ii.pdelta, 
                                                       lat = tr_ii.lat ) 
        
        data['pressure'] = (['time','y','x'], pressure)
        data['pressure'].attrs = {'long_name':'Surface atmospheric pressure',
                                  'units':'millibar'}
        data['windspeed'] = (['time','y','x'], wind_g)
        data['windspeed'].attrs = {'long_name':'Magnitude of windspeed',
                                   'units':'m/s'}
        data.attrs['tc_model'] = 'Holland1980'
        data.attrs['tc_B_model'] = B_model

        # Expand windspeed into wind vectors with 0 inflow angle
        self.apply_inflow_angle( inflow_model = 'constant', inflow_angle = 0)

    @classmethod
    def pressure_equation(cls, dist_cent, rmw, B,
                          penv, pcen, lat ):
        """Tropical cyclone pressure model taken from (Holland, 1980).
        This models the pressure profile as a hyperbolic function of
        distance from storm center, radius of max winds, a shape parameter
        B, central pressure and latitude. The shape parameter B can be 
        estimated using one of the B_* functions inside this class.

        This function will calculate pressure for a single snapshot of a 
        tropical cyclone track, i.e. all track parameters should be floats.

        Args:
            dist_cent (np.ndarray): An array of distances to the storm center (km).
                This array can be any shape, e.g. a 1D profile or a 2D grid.
            rmw (float): Radius of maximum winds (km)
            B (float): The storm shape parameter (dimensionless). 
            penv (float): Environmental pressure (mb)
            pcen (float): Central pressure (mb)
            lat (float): Latitude (degrees)
    
        Returns:
            Pressure array of the same shape as dist_cent, units are mb.
        """

        # Convert units for the Holland equation
        penv = _const.mb_to_pa * penv
        pcen = _const.mb_to_pa * pcen
        dist_cent = _const.km_to_m * dist_cent
        rmw = _const.km_to_m * rmw

        # Pressure equation
        rmw_norm = (rmw / dist_cent)**B
        pressure = pcen + (penv - pcen)*np.exp( -rmw_norm )

        # Get locations where distances are very close to eye (approaches infinity)
        too_close = dist_cent < 1
        pressure[too_close] = pcen
        return pressure / _const.mb_to_pa
    
    @classmethod
    def gradient_wind_equation(cls, dist_cent, rmw, B, pdelta, lat ):
        """Tropical cyclone gradient wind model taken from (Holland, 1980).
        Calculates 1-min wind speeds at gradient level.

        This function will calculate pressure for a single snapshot of a 
        tropical cyclone track, i.e. all track parameters should be floats.

        Args:
            dist_cent (np.ndarray): An array of distances to the storm center (km).
                This array can be any shape, e.g. a 1D profile or a 2D grid.
            rmw (float): Radius of maximum winds (km)
            B (float): The storm shape parameter (dimensionless). 
            penv (float): Environmental pressure (mb)
            pcen (float): Central pressure (mb)
            lat (float): Latitude (degrees)
    
        Returns:
            Pressure array of the same shape as dist_cent, units are mb.
        """
        # Convert units
        pdelta = _const.mb_to_pa * pdelta
        dist_cent = _const.km_to_m * dist_cent
        rmw = _const.km_to_m * rmw
        
        # Calculate the various terms one by one
        f = _utils.calculate_coriolis( lat )
        rf = .5 * dist_cent * f
        rf2 = rf**2
        rmw_norm = (rmw / dist_cent)**B
    
        inside_sqrt = rmw_norm * (B / _const.rho) * pdelta * np.exp( -rmw_norm) + rf2
        Vg = np.sqrt( np.fmax(0, inside_sqrt) ) - rf
        Vg[ dist_cent < 0.1 ] = 0
        return Vg

    @classmethod
    def B_powell05(cls, rmw, lat ):
        ''' Statistical model of Holland B parameter according to (Powell, 2005).
        Returns B in same form as input arguments.
        
        Args:
            rmw (float): Radius of maximum winds (km)
            lat (float): Latitude (degrees)
        '''
        return 1.881 - 0.00557*rmw - 0.01295*lat

    @classmethod
    def B_wr04(cls, rmw, vmax, lat ):
        ''' Statistical model of Holland B parameter according to 
            (Willoughby & Rahn, 2004). Returns B in same form as input arguments.
        
        Args:
            rmw (float, np.ndarray): Radius of maximum winds (km)
            vmax (float, np.ndarray): Maximum sustained wind at rmax (ms-1)
            lat (float, np.ndarray): Latitude (degrees)
        '''
        B = 1.0036 + 0.0173*vmax + 0.0313*np.log(rmw) + 0.0087*lat
        return B 
                       
    @classmethod
    def B_vickery00(cls, pdelta, rmw ):
        ''' Statistical model of Holland B parameter according to (Vickery, 2000).
        Returns B in same form as input arguments.
        
        Args:
            pdelta (float, np.ndarray): Difference between environmental and central
                pressure (mb).
            rmw (float, np.ndarray): Radius of maximum winds (km)
        '''
        B = 1.38 - 0.00184*pdelta + 0.00309*rmw
        return B

    @classmethod 
    def B_holland80(cls, pdelta, vmax ):
        ''' Statistical model of Holland B parameter according to (Holland, 1980).
        Returns B in same form as input arguments.
        
        Args:
            pdelta (float, np.ndarray): Difference between environmental and central
                pressure (mb).
            vmax (float, np.ndarray): Maximum sustained wind at rmax (ms-1).
        '''
        # Convert units
        pdelta = pdelta * _const.mb_to_pa
        
        B = vmax**2 * np.exp(1) * _const.rho /  pdelta
        return B
    



import numpy as np

def calculate_rmw( track, rmw_model = 'vickery08' ):
    ''' Calculate rmw from track dataframe

    Args:
        track (pd.Dataframe): Track dataframe with columns necessary for chosen rmw models
        rmw_model (str): rmw model to use from rmw_models module.

    Returns:
        rmw (np.ndarray): Radius of maximum winds in km
    '''

    if rmw_model == 'vickery08':
        rmw = vickery08( track.pdelta, track.lat )
    elif rmw_model == 'climada':
        rmw = rmw_climada( track.pcen )

    return rmw

def vickery08( pdelta, lat ):
    ''' Statistical model of rmw, taken from (Vickery & Wadhera, 2008).
    This is for all hurricanes analysed. Returns rmw in km for pdelta in mb.'''
    exponent = 3.015 - 6.291e-5*pdelta**2 + 0.0337*lat
    return np.exp(exponent)

def rmw_climada(pcen):
    """
    Statistical model for rmw (km), as in the Climada package.
    """
    pres_l = [872, 940, 980, 1021]
    rmw_l = [14.907318, 15.726927, 25.742142, 56.856522]
    rmw = np.zeros_like
    for i, pres_l_i in enumerate(pres_l):
        slope_0 = 1. / (pres_l_i - pres_l[i - 1]) if i > 0 else 0
        slope_1 = 1. / (pres_l[i + 1] - pres_l_i) if i + 1 < len(pres_l) else 0
        rmw[msk] += rmw_l[i] * np.fmax(0, (1 - slope_0 * np.fmax(0, pres_l_i - pcen[msk])
                                           - slope_1 * np.fmax(0, pcen[msk] - pres_l_i)))

    return np.where(rmw <= 0, np.nan, rmw)


import numpy as np
from paratc import _utils, _const
import pandas as pd
from shapely.geometry import LineString, Point, Polygon
from datetime import datetime, timedelta

def filter_bad_rmw( track_list ):

    output_list = []

    for ii, tr in enumerate(track_list):
        zero_sum = np.sum( tr.radius_max_wind.values != 0 )
        nan_sum = np.sum( ~np.isnan( tr.radius_max_wind.values ) )
        if zero_sum > 0 and nan_sum > 0:
            output_list.append( tr )
    return output_list

def interpolate_to_timestep( track, new_timestep, **kwargs):
    ''' Interpolate a track dataframe to a new timestep in hours '''

    time_0 = track.time.values
    new_time = pd.date_range( time_0[0], time_0[-1], freq=f'{new_timestep}H')

    # Convert to xarray for easier interpolation
    track = track.to_xarray().swap_dims({'index':'time'})
    track = track.interp( time = new_time, **kwargs )
    track = track.to_dataframe().reset_index()
    track['timestep'] = new_timestep
    return track.drop(columns='index')
    

def get_translation_vector( track_lon, track_lat, track_timestep ):
    ''' Get storm translation vectors from track longitudes, latitudes and
    timestep. 

    Args:
        track_lon (np.ndarray):
        track_lat (np.ndarray):
        track_timestep (np.ndarray):
    '''

    utrans = np.zeros_like(track_lon)
    vtrans = np.zeros_like(track_lat)
    n_steps = len(track_lon)
    
    distances = np.zeros_like(track_lon)
    distances[1:] = [ _utils.haversine( track_lon[ii], track_lat[ii], 
                                        track_lon[ii-1], track_lat[ii-1], radians=False ) 
                                        for ii in range(1,n_steps) ]
    trans_speed = _const.kmh_to_ms * distances / track_timestep

    utrans[1:] = track_lon[1:] - track_lon[:-1]
    vtrans[1:] = track_lat[1:] - track_lat[:-1]
    vec_norm = np.sqrt( utrans**2 + vtrans**2 )
    utrans = trans_speed * utrans / vec_norm
    vtrans = trans_speed * vtrans / vec_norm
    utrans[0] = 0
    vtrans[0] = 0
    utrans[trans_speed == 0] = 0
    vtrans[trans_speed == 0] = 0
    
    return trans_speed, utrans, vtrans

def filter_tracks_by_column( track_list, col_name = 'vmax',
                             col_min = 33, col_max = np.inf):
    ''' Filter a list of track dataframes by a range of values in a specified
    column. By default, this will filter all tracks that never reach 64 m/s.
    
    Args:
        track_list (list): List of dataframes containing track info.
                           Must contain vmax column.
        col_name (str): Name of column to filter by (default = 'vmax')
        col_min (float): Minimum value of variable to search
        col_max (float): Maximum value of variable to search

    Returns:
        New filtered list of track dataframes.
    '''
    keep_idx = []
    for ii, tr in enumerate(track_list):
        var_over = tr[col_name].values > col_min
        var_under = tr[col_name].values < col_max
        if np.sum( np.logical_and( var_over, var_under ) ) > 0:
            keep_idx.append(ii)
    return keep_idx

def distance_track_to_poly( track_list, pol ):
    ''' Uses Shapely to check minimum proximity of a storm track to a polygon.
        This does not calculate geographical distances, result will be in 
        degrees. This function is used for determining if a track passes through
        a polygon.'''
    linestrings = track_to_linestring( track_list )
    return pol.distance(linestrings)

def clip_track_to_poly( track, poly, max_dist = 1, round_days=True ):
    '''
    Takes a track dataframe and clips it to a shapely polygon.

    Args:
        track (pd.dataframe): Pandas dataframe track with lon, lat columns
        poly (shapely Polygon): Poly to clip to in same crs
        max_dist (float): Maximum distance around poly to clip (degrees)
        round_days (bool): If true, round resulting poly to day start
    '''
    t_points = list(zip( track.lon, track.lat) )
    points = [Point(tc) for tc in t_points]
    dist = np.array( [poly.distance(pt) for pt in points] )
    keep_idx = np.where(dist <= max_dist)[0]
    track_clipped = track.iloc[np.min(keep_idx):np.max(keep_idx)]

    if round_days: 
        date0 = datetime(*pd.to_datetime(track_clipped.time.values[0]).timetuple()[:3])
        date1 = datetime(*pd.to_datetime(track_clipped.time.values[-1]).timetuple()[:3])
        date1 = date1 + timedelta(days=1)
        didx = np.logical_and( track.time >= date0, track.time <= date1 )
        track_clipped = track.iloc[ np.where(didx)[0]]

    return track_clipped

def track_to_linestring( track ):
    ''' Converts a track (or list of) dataframes into a list of shapely (lon,lat) LineStrings '''
    if type(track) is not list:
        track = [track]
    ls_list =[ LineString(list(zip( tr.lon.values, tr.lat.values) ) )  for tr in track ]

    if len(track) > 1:
        return ls_list
    else:
        return ls_list[0]

def subset_tracks_in_poly( track_list, pol, buffer = 0):
    ''' Identifies tracks (dataframe) in a list that pass through a specified
    polygon, with some added buffer. This is not an exact procedure and geographical distances
    are not used for the buffer. Instead, distance in degrees is used.
    
    Args:
        track_list (list): List of track pandas dataframes with lon, lat columns
        pol (shapely.geometry.Polygon): Polygon to compare
        buffer (float): Buffer around polygon to allow (in degrees)
    returns
        Indices of filtered tracks.
    '''
    distances = distance_track_to_poly( track_list, pol )
    keep_idx = np.where( distances <= buffer )[0]
    return keep_idx

def climada_to_dataframe( track, convert_units = True ):
    ''' Converts a climada track xarray dataset into an appropriate dataframe for ParaTC'''
    df_track = track.rename({'radius_max_wind':'rmw', 
                             'central_pressure':'pcen',
                             'environmental_pressure':'penv',
                             'max_sustained_wind':'vmax'})
    df_track = df_track.to_dataframe().reset_index()
    if convert_units:
        df_track['rmw'] = df_track['rmw']*1.852
        df_track['vmax'] = df_track['vmax']*0.51444 / 0.9
    return df_track

def subset_tracks_in_year( track_list, year ):
    ''' Subsets tracks into integer year '''
    year_list = [ pd.to_datetime(tr.time[0]).year for tr in track_list ]
    year_list = np.array(year_list)
    keep_idx = np.where(year_list == year)[0]
    return keep_idx


import numpy as np
from paratc import tctools, inflow_models, bg_models, stress_models, rmw_models
from paratc import _utils, _const, _output_formats, track_tools
import xarray as xr
import os
import matplotlib.pyplot as plt

class TCModel():

    ''' General tropical cyclone model. This is inherited by tropical cyclone model classes such as
    Holland1980(). This class contains more general functions that can be applied to any storm model
    dataset. This will calculate some variables if they are not already in track dataset:

        pdelta: Difference between central pressure and environmental pressure
        timestep: Timestep in hours between subsequent track steps
        utrans, vtrans: Storm translation vectors in m/s
        rmw: Radius of max winds, generated using a statistical model where not present.

    Additionally, rmw will be checked for missing values (0 or NaN) and have them filled with
    statistical estimates. Some functions will require vmax (maximum wind speed), which should be
    in m/s.

    Args:
        track (pd.DataFrame): Pandas Dataframe with columns representing tropical cyclone track.
        grid_lon (np.ndarray): 2D grid longitudes
        grid_lat (np.ndarray): 2D grid latitudes
        rmw_model (np.ndarray): Radius of max winds model to fill missing RMW values.
                                default = 'vickery08'. See rmw_models.py for more info.
        subtract_trans_speed (bool): If true, translation speed will be subtracted from
                                     vmax in the track dataframe.
        interp_timestep (float): If specified, track will be interpolated to a new timestep (hours).
    '''
    def __init__(self, track, grid_lon, grid_lat, 
                       rmw_model = 'vickery08',
                       subtract_trans_speed = True,
                       interp_timestep = None):

        if 'time' not in track:
            raise Exception(" time column not found in track dataframe.")

        track = track.copy()

        # Where (if) central pressure > env pressure then set to env pressure
        c_gt_env = track['pcen'] > track['penv']
        track['pcen'] = track['pcen'].where( ~c_gt_env, track['penv'])
        if 'pdelta' not in track:
            track['pdelta'] = track['penv'] - track['pcen']

        # Calculate RMW and replace missing values (or missing column..)
        rmw = rmw_models.calculate_rmw( track, rmw_model = rmw_model )
        if 'rmw' not in track:
            track['rmw'] = rmw
        else:
            track['rmw'] = track.rmw.where( track.rmw != 0, rmw )
            track['rmw'] = track.rmw.where( ~np.isnan(track.rmw), rmw )

        # Do interpolation of time if wanted
        if interp_timestep is not None:
            track = track_tools.interpolate_to_timestep( track, interp_timestep )

        # Get hour and timestep
        if 'timestep' not in track:
            track['timestep'] = _utils.get_timestep_from_time( track.time )
        if 'hour' not in track:
            track['hour'] = np.cumsum( track['timestep'] ) - track['timestep'][0]
        n_time = len(track)

        # Get translation velocity
        if 'utrans' not in track or 'vtrans' not in track:
            trans_speed, utrans, vtrans = track_tools.get_translation_vector( track.lon.values, 
                                                             track.lat.values, 
                                                             track.timestep.values )
            track['utrans'], track['vtrans'] = utrans, vtrans
            track['trans_speed'] = trans_speed

        # If vmax, subtract translation speed
        if subtract_trans_speed and 'vmax' in track:
            track['vmax'] = track['vmax'] - track['trans_speed']
        
        # Calculate distances from storm center
        dist_cent = np.zeros( (n_time, *grid_lon.shape) )
        for ii in range( n_time ):
            dist_cent[ii] = _utils.haversine( grid_lon, grid_lat, 
                                              track.iloc[ii].lon, 
                                              track.iloc[ii].lat, 
                                              radians=False )

        # Make output grid dataset and save to object
        data = xr.Dataset()
        if 'time' in track:
            data['time'] = (['time'], track.time)
        else:
            data['time'] = (['time'], np.arange(n_time))

        data['lon'] = (['y','x'], grid_lon)
        data['lat'] = (['y','x'], grid_lat)
        data['dist_cent'] = (['time','y','x'], dist_cent)
        data['dist_cent'].attrs = {'long_name':'Distance from storm center',
                                   'units':'km'}
        data = data.set_coords(['lon','lat','time'])

        # Assign variables to this instance
        self.data = data
        self.n_time = len(track)
        self.track = track
        self.hemisphere = _utils.get_hemisphere( track.lat[0] )

    def scale_winds( self, alpha ):
        '''
        Scales windspeeds and wind vectors as alpha * windspeed, alpha * wind_u, alpha * wind_v.
        This is done in place.

        You may want to scale winds to e.g:
            1. Convert from gradient level to surface level winds
            2. Convert from 1-min to 10-min sustained winds.

        Args:
            alpha (float): Scaling parameter
        '''
        data = self.data
        data['windspeed'] = data['windspeed'] * alpha
        data['wind_u'] = data['wind_u'] * alpha
        data['wind_v'] = data['wind_v'] * alpha
        

    def apply_inflow_angle(self, inflow_model = 'constant', inflow_angle = 0):
        '''
        Make wind U and V components using the combination of windspeed and an
        inflow angle model. These components will be saved to the object's dataset using
        names wind_u and wind_v.

        Args:
            inflow_model (str): Inflow angle model to use. Default is 'constant', which will
                apply a constant rotate to all vectors. Other options:
                    nws: Apply piecewise function of distance. See inflow_models.nws() for more.
                    constant: Applies a constant inflow angle of inflow_angle degrees.
            inflow_angle (str): If inflow_angle is 'constant', this is the angle to apply.
        '''

        # Expand wind speed into vectors using inflow angle
        n_time = len(self.track)
        data = self.data
        wind_u = np.zeros( (n_time, *data.lon.shape) )
        wind_v = np.zeros( (n_time, *data.lon.shape) )

        for tii in range( n_time ):

            # Get relevant variables for this iteration
            ws_ii = data.windspeed[tii].values
            dist_ii = data.dist_cent[tii].values
            tr_ii = self.track.iloc[tii]
        
            # Make inflow angle
            if inflow_model == 'nws':
                inflow_angle = inflow_models.nws( dist_ii, tr_ii.rmw )
            elif inflow_model == 'constant': 
                # In this case we will just use inflow_angle
                pass
            else:
                raise Exception(f' Inflow angle model not found: {inflow_model}' )

            # Apply inflow angle and save to output array
            u_ii, v_ii = tctools.windspeed_to_vector( ws_ii, inflow_angle, 
                                                      data.lon, data.lat,
                                                      tr_ii.lon, tr_ii.lat )
            wind_u[tii] = u_ii
            wind_v[tii] = v_ii

        # Save data to object's dataset
        data['wind_u'] = (['time','y','x'], wind_u)
        data['wind_u'].attrs = {'long_name':'U-component (east) of wind vector',
                                'units':'m/s'}
        data['wind_v'] = (['time','y','x'], wind_v)
        data['wind_v'].attrs = {'long_name':'V-component (north) of wind vector',
                                'units':'m/s'}
        data.attrs['inflow_model'] = inflow_model
        if inflow_model == 'constant':
            data.attrs['inflow_angle'] = inflow_angle

    def add_background_winds(self, bg_model='constant', 
                             bg_alpha=.55, bg_beta=20):
        ''' 
        Add background wind field to the cyclone using a specified model, rotation and scaling.
        This step boils down to adding some form of the environmental flow to the model,
        which helps create the necessary asymmetries in the cyclone model. Using this function
        modifies the model in place, making changes to wind_u and wind_v. Windspeed will also
        be recalculated.

        When bg_model == 'constant', the storm translation speed is applied constantly
        throughout the domain. Otherwise, a model (e.g. radial) is used. see the bg_models.py
        module for functions.

        Args:
            bg_model (str): The background flow model to add to our storm. Options:
                'constant': Adds storm translation vector to cyclone uniformly.
                'MN05': Adds translation vector multiplied by a reciprocal distance 
                    decay function according to Mouton & Nordbeck (2005).
                'miyazaki61': Add translation vector multiplied by exponential 
                    distance decay function, according to Miyazaki, (1961)
            bg_alpha (float): The scaling to multiple background winds before adding to storm
                e.g. to bring down to surface level. Default = .55
            bg_beta (float): Rotation to apply to background wind fields before adding to 
                storm. Default = 20.
        '''

        # Get data and loop over track timesteps
        data = self.data
        wind_u = data.wind_u
        wind_v = data.wind_v

        for tii in range(self.n_time):

            tr_ii = self.track.iloc[tii]
            
            # Make background flow
            hemisphere = _utils.get_hemisphere( self.track.lat[0] )
            if bg_model == 'constant':
                u_bg, v_bg = tr_ii.utrans, tr_ii.vtrans
            elif bg_model == 'miyazaki61':
                u_bg, v_bg = bg_models.miyazaki61( data.dist_cent[tii].values, 
                                                   tr_ii.utrans, tr_ii.vtrans, 
                                                   tr_ii.rmw )
            elif bg_model == 'MN05':
                u_bg, v_bg = bg_models.MN05( data.dist_cent[tii].values, 
                                             tr_ii.utrans, tr_ii.vtrans, 
                                             tr_ii.rmw )
            else:
                raise Exception(f' Background flow model not found: {bg_model}' )

            if bg_beta != 0:
                if self.hemisphere == 'N':
                    u_bg, v_bg = _utils.rotate_vectors( u_bg, v_bg, bg_beta, radians=False)
                else:
                    u_bg, v_bg = _utils.rotate_vectors( u_bg, v_bg, -bg_beta, radians=False)
            wind_u[tii] += bg_alpha * u_bg
            wind_v[tii] += bg_alpha * v_bg

        # Reconstruct windspeed from vectors
        self.calculate_windspeed_from_uv()

        # Save attributes
        data.attrs['bg_model'] = bg_model
        data.attrs['bg_alpha'] = bg_alpha
        data.attrs['bg_beta'] = bg_beta

    def make_wind_stress( self, cd_model='large_pond', cd_min = 0, cd_max = 3.5e-3 ):
        '''
        Calculates wind stress as a quadratic function of windspeed, applying a 
        statistical model of the drag coefficient C_d. 

        Windstress vectors will be saved to the TCModel.data dataset as 
        stress_u and stress_v.
    
        Args:
            cd_model (str): The drag coefficient model to use. Options:
                'andreas12': According to Andreas et al., (2012)
                'large_pond82': According to Large & Pond (1982)
                'garratt77': According to Garratt (1977)
                'peng_li15': According to Peng & Li (2015)
                Standalone drag functions can be found in stress_models.py
            cd_min (float): Minimum value for drag coefficient CD
            cd_max (float): Maximum value for drag coefficient CD
        '''

        data = self.data
        
        # Calculate CD according to model and clip to min, max bounds
        if cd_model == 'andreas12':
            cd = stress_models.cd_andreas( data.windspeed )
        elif cd_model == 'large_pond82':
            cd = stress_models.cd_large_pond82( data.windspeed )
        elif cd_model == 'garratt77':
            cd = stress_models.cd_garratt77( data.windspeed )
        elif cd_model == 'peng_li15':
            cd = stress_models.cd_peng_li15( data.windspeed )
        else:
            raise Exception(f' Unknown drag coefficient model: {cd_model}')
    
        cd = np.clip(cd, cd_min, cd_max)

        # Calculate wind stress vectors
        tau, tau_u, tau_v = stress_models.quadratic_stress_equation( data.wind_u, 
                                                                     data.wind_v,
                                                                     cd = cd )

        # Save to dataset
        data['stress_u'] = (['time','y','x'], tau_u.values)
        data['stress_v'] = (['time','y','x'], tau_v.values)
        data.attrs['cd_model'] = cd_model
        data.attrs['cd_min'] = cd_min
        data.attrs['cd_max'] = cd_max
        

    def calculate_windspeed_from_uv(self):
        ''' Uses pythagoras to recalculate windspeed from U and V vectors within this object '''
        windspeed = np.sqrt( self.data.wind_u**2 + self.data.wind_v**2 )
        self.data['windspeed'] = windspeed
    
    def interpolate_to_arakawa(self, lon_u, lat_u, lon_v, lat_v):
        ''' Interpolate the current tropical cyclone onto an Arakawa rho, U, V grid.
        This function uses XESMF to interpolate the dataset in space. Assumes your current
        storm is already at density points.
        Make sure your python version is >= 3.9.
        '''

        import xesmf as xe
        data = self.data

        # Find lists of variables present in dataset
        u_vars = [ vname for vname in ['stress_u','wind_u'] if vname in data ]
        v_vars = [ vname for vname in ['stress_v','wind_v'] if vname in data ]
    
        # Regrid stresses onto U using XESMF
        ds_u = xr.Dataset( coords = dict( lon = (['y','x'], lon_u),
                                          lat = (['y','x'], lat_u) ) )
        regridder = xe.Regridder(data, ds_u, "bilinear")
        ds_u = regridder( data[u_vars] )
        ds_u = ds_u.rename({'lon':'lon_u', 'lat':'lat_u'})
        ds_u = ds_u.swap_dims({'x':'x_u', 'y':'y_u'})

        # Regrid stresses onto V using XESMF
        ds_v = xr.Dataset( coords = dict( lon = (['y','x'], lon_v),
                                          lat = (['y','x'], lat_v) ) )
        regridder = xe.Regridder(data, ds_v, "bilinear")
        ds_v = regridder( data[v_vars] )
        ds_v = ds_v.rename({'lon':'lon_v', 'lat':'lat_v'})
        ds_v = ds_v.swap_dims({'x':'x_v', 'y':'y_v'})

        # Place interpolated variables back into self.data
        for vname in u_vars:
            data[vname] = ds_u[vname]
        for vname in v_vars:
            data[vname] = ds_v[vname]
            
    def to_ROMS(self, ds_grd, rotate_to_grid = True ):
        '''
        Convert storm dataset to a format appropriate for ROMS forcing format.
        This will be done in place, changing self.data.

        Dataset will be interpolated onto U and V coordinates using interpolate_to_arakawa(),
        which uses XESMF. Variables will be renamed and relevant attributes will be added.
        The storm's current lon and lat variables will be mapped directly onto rho points.
        Make sure your python version is >= 3.9.

        Args:
            ds_grd (xr.Dataset): ROMS grid in xarray Dataset form, opened from netcdf.
                Must contain rho, u and v coordinates.
            rotate_to_grid (bool): If true, rotate vector variables to be relative to
                ROMS grid, rather than North/East. This is done by applying an anti-clockwise
                rotation using ds_grd.angle.
        '''

        data = self.data

        # Rotate grid using rotate_vectors()
        if rotate_to_grid:
            rot_angle = ds_grd.angle.rename({'eta_rho':'y', 'xi_rho':'x'})
            U_rot, V_rot = _utils.rotate_vectors( self.data['stress_u'],
                                                  self.data['stress_v'],
                                                  -rot_angle, radians = True )
            data['stress_u'] = U_rot
            data['stress_v'] = V_rot
        
        # Interpolate to Arakawa grid first
        self.interpolate_to_arakawa( ds_grd.lon_u.values, ds_grd.lat_u.values,
                                     ds_grd.lon_v.values, ds_grd.lat_v.values )
        
        # Rename and assign attributes
        if 'stress_u' in data:
            self.data = _output_formats.make_ROMS_dataset( ds_grd, time = self.track.time, 
                                                   stress_u = data.stress_u.values, 
                                                   stress_v = data.stress_v.values, 
                                                   pressure = data.pressure.values )
        else:
            self.data = _output_formats.make_ROMS_dataset( ds_grd, time = self.track.time, 
                                                   pressure = data.pressure.values )

        return

    def to_netcdf( self, fp_out, **kwargs ):
        ''' Write this objects dataset to netCDF file using xarray.Dataset.to_netcdf().
        Kwargs are passed to the xarray function. Will remove file if it already exists.

        Args:
            fp_out (str): Output file string.
        '''

        if os.path.exists(fp_out):
            os.remove(fp_out)
        self.data.to_netcdf( fp_out, **kwargs )

    def plot( self, timestep = None, field = 'windspeed' ):
        ''' Quick plotting routined to view the generated storm. 

        Args:
            timestep (int, None): If None (default), will plot the envelope of field.
                If int, will plot the storm for that timestep
            field (str): Field to plot. Either 'pressure' or 'windspeed'
        '''
        
        data = self.data
        f = plt.figure( figsize = (6,5))

        # Set cmap and title
        if field == 'windspeed':
            cmap = plt.get_cmap('Reds',8)
            title = 'Parametric TC wind speed (m/s)'
        elif field == 'pressure':
            cmap = plt.get_cmap('Greens_r',8)
            title = 'Parametric TC pressure (m/s)'

        if timestep is not None:
            plt.pcolormesh( data.lon, data.lat, data[field].isel(time=timestep), cmap = cmap)
            plt.colorbar()
            plt.quiver( data.lon[::3,::3], data.lat[::3,::3], 
                        data.wind_u[timestep][::3,::3], 
                        data.wind_v[timestep][::3,::3] )
        else:
            if field == 'windspeed':
                data_opt = data.windspeed.max(dim='time')
            elif field == 'pressure':
                data_opt = data.pressure.min(dim='time')
            plt.pcolormesh( data.lon, data.lat, data_opt, cmap = cmap ) 

        plt.plot( self.track.lon, self.track.lat, c='b', linewidth=1 )
        plt.scatter( self.track.lon, self.track.lat, c='b', s=10)

        plt.xlim( data.lon.min(), data.lon.max())
        plt.ylim( data.lat.min(), data.lat.max())
        plt.title(title)

        


'''
Functions for calculating surface wind stress. This module contains functions for both
the calculate of stress and the crag coefficients C_D.
'''

import numpy as np
from paratc import _const

def quadratic_stress_equation( wind_u = None, wind_v = None, windspeed = None, 
                               cd = 2.5e-3 ):
    '''
    Calculates wind stress as a quadratic function of windspeed, applying a 
    statistical model of the drag coefficient C_d. You must provide either windspeed
    or both of (wind_u, wind_v) to this function.

    Args:
        wind_u (float, np.ndarray): Wind U vector components (ms^-1)
        wind_v (float, np.ndarray): Wind V vector components (ms^-1)
        windspeed (float, np.ndarray): Windspeed (ms^-1)
        cd (float): The drag coefficient cd.
    Returns:
        Windstress (tau) if only windspeed is provided. If wind_u and wind_v are provided,
        also return tau_u and tau_v components of stress
    '''

    if windspeed is None:
        windspeed= np.sqrt(wind_u**2 + wind_v**2)

    tau = windspeed**2 * _const.rho * cd

    if wind_u is not None and wind_v is not None:
        tau_u = wind_u * windspeed * _const.rho * cd
        tau_v = wind_v * windspeed * _const.rho * cd
        return tau, tau_u, tau_v
    else:
        return tau

def cd_garratt77( windspeed ):
    ''' Wind stress drag coefficient according to Garratt, (1977) '''
    cd = 0.001*(0.75+0.067*windspeed)
    return cd
    
def cd_large_pond82( windspeed ):
    ''' Wind stress drag coefficient according to Large & Pond (1982). 
    This is not quite what is presented in the paper. We also hold cd = 1.14e-3
    for windspeeds < 4 and cd = 2.18e-3 for windspeed >= 26.'''

    cd = (0.49+0.065*windspeed) * 0.001
    return np.clip(cd, 1.14e-3, (0.49+0.065*26) * 0.001)

def cd_andreas12( windspeed ):
    ''' Wind stress drag coefficient according to Andreas et al., (2012) '''
    return 3.4e-3 * (1 - 4.17 / windspeed )**2

def cd_peng_li15( windspeed ):
    ''' Wind stress drag coefficient according to Peng & Li (2015) '''
    a = 2.15e-6
    c = 2.797e-3
    cd = -a*(windspeed - 33)**2 + c
    return cd


'''
Models of inflow angle.
'''

import numpy as np

def nws( dist_cent, rmw ):
    '''
    Inflow angle determined using a piecewise function of storm center distance,
    as recommended by the National Weather Service (Bretschnedier, 1972)

    Args:
        dist_cent (float, np.ndarray): Distance from storm center (km).
        rmw (float): radius of maximum winds (km)

    Returns:
        inflow_angle( float, np.ndarray): Angles in degrees, in same format as dist_cent.
    '''
    inflow_angle = np.zeros_like(dist_cent)
    r_scaled = dist_cent / rmw

    # Close zone
    zone_angle = 10*(r_scaled + 1)
    zone_idx = dist_cent <= rmw
    inflow_angle[ zone_idx ] = zone_angle[zone_idx]

    # Middle zone
    zone_angle = 25*r_scaled - 5
    zone_idx = np.logical_and( dist_cent >= rmw, 
                               dist_cent <= 1.2*rmw )
    inflow_angle[ zone_idx ] = zone_angle[zone_idx]

    # Far zone
    inflow_angle[ dist_cent >= 1.2*rmw ] = 25
    return inflow_angle


def make_ROMS_dataset( ds_grd, time, 
                       stress_u = None, stress_v=None, 
                       pressure = None, 
                       track_lon = None, track_lat = None):
    ''' Create a forcing dataset in the correct format for ROMS. You can provide
    each oof pressure, stress_u and stress_v optionally. Only those provided will
    be saved to the dataset. Track_lon and track_lat do not change the dataset, but
    can be optionally saved to the dataset for later reference.

    Args:
        ds_grd (xr.Dataset): ROMS grid file, read into an xarray Dataset
        time (array ): Time variable for dataset.
        stress_u (np.ndarray): U component of windstress with dimension (time, y, x)
        stress_v (np.ndarray): V component of windstress with dimension (time, y, x)
        pressure (np.ndarray): Surface pressure with dimension (time, y, x)
        track_lon (np.ndarray): Tropical cyclone track longitudes
        track_lat (np.ndarray): Tropical cyclone track latitudes.

    Returns:
        Xarray Dataset object in ROMs format, ready for writing.
        
    '''

    ds_tmp = ds_grd[['lon_rho','lat_rho','lon_u','lat_u','lon_v','lat_v']]
    ds_tmp['sms_time'] = (['sms_time'], time)
    ds_tmp['pair_time'] = (['pair_time'], time)

    if stress_u is not None:
        ds_tmp['sustr'] = (['sms_time','eta_u', 'xi_u'], stress_u)
        ds_tmp['sustr'].attrs = {'long_name':"surface u-momentum stress",
                                 'units':"m/s",
                                 'field':"surface u-momentum stress",
                                 'time':"sms_time",
                                 'coordinates':'lon_u lat_u' }
    if stress_v is not None:
        ds_tmp['svstr'] = (['sms_time','eta_v', 'xi_v'], stress_v)
        ds_tmp['svstr'].attrs = {'long_name':"surface v-momentum stress",
                                 'units':"Newton meter-2",
                                 'field':"surface v-momentum stress",
                                 'time':"sms_time",
                                 'coordinates':'lon_v lat_v' }

    if pressure is not None:
        ds_tmp['Pair'] = (['pair_time','eta_rho', 'xi_rho'], pressure)
        ds_tmp['Pair'].attrs = {'long_name':"surface air pressure",
                                 'units':"millibar",
                                 'field':"surface air pressure",
                                 'time':"pair_time",
                                 'coordinates':'lon_rho lat_rho' }
        
    ds_tmp.sms_time.encoding['units'] = 'days since 1900-01-01'
    ds_tmp.pair_time.encoding['units'] = 'days since 1900-01-01'

    if track_lon is not None:
        ds_tmp['track_lon'] = (['sms_time'], track_lon)
        ds_tmp['track_lat'] = (['sms_time'], track_lat)

    return ds_tmp
